SystemOrganization addCategory: #'Rest-Server'!

HttpService subclass: #RestHttpService
	instanceVariableNames: 'handlers'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Rest-Server'!

RestHttpService class
	instanceVariableNames: 'service'!

!RestHttpService commentStamp: 'mt 2/16/2009 14:56' prior: 0!
I am a restful http service. I am primarily used for the XP integration platform project. (Package: 'Xpf')!

RestHttpService class
	instanceVariableNames: 'service'!

!RestHttpService class methodsFor: 'service' stamp: 'mt 2/16/2009 15:14'!
isRunning
	^service isNil not and: [service isRunning]! !

!RestHttpService class methodsFor: 'service' stamp: 'mt 2/16/2009 14:57'!
port
	^8082! !

!RestHttpService class methodsFor: 'service' stamp: 'mt 2/16/2009 15:10'!
service
	^service! !

!RestHttpService class methodsFor: 'service' stamp: 'mt 2/16/2009 15:04'!
start
	"Start exactly one http service on port 8082. This service will be closed first if you call this message twice."

	self stop.
	service := (RestHttpService on: self port)
		start;
		yourself.! !

!RestHttpService class methodsFor: 'service' stamp: 'mt 2/16/2009 14:52'!
stop
	"Shutdown the server. Does nothing if server is not running."

	service ifNotNil:
		[service unregister.
		service := nil].! !

!RestHttpService methodsFor: 'initialization' stamp: 'mt 2/2/2009 10:12'!
initialize
	super initialize.
	handlers := OrderedCollection new	"needed for priorities"! !

!RestHttpService methodsFor: 'serving' stamp: 'cs 2/15/2009 16:07'!
installHandler: aRequestHandlerClass on: aPathString
	(handlers anySatisfy: [:ass | ass key = aPathString ])
		ifTrue: [^ false].
	handlers add: aPathString -> aRequestHandlerClass! !

!RestHttpService methodsFor: 'serving' stamp: 'mt 2/16/2009 18:02'!
processHttpRequest: request
	
	| activeHandler path |
	request isPutRequest ifTrue: [Transcript show: request fields].
	activeHandler := (handlers detect: [:ass| 
		path := ass key.
		request url includesSubString: path
		]) value.
	^(activeHandler new) process: request! !

Object subclass: #RestHttpHandler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Rest-Server'!

!RestHttpHandler methodsFor: 'request processing' stamp: 'mt 2/16/2009 15:18'!
process: anHttpRequest
	"Handles an http request. Should answer with an http response."
	
	self subclassResponsibility! !

RestHttpHandler subclass: #RestPerformHttpHandler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Rest-Server'!

!RestPerformHttpHandler methodsFor: 'request processing' stamp: 'mt 2/16/2009 15:25'!
process: anHttpRequest
	"Takes first element in path as classname and sends message from third path element to that class."

	|pc class selector result|
	pc := (URI fromString: anHttpRequest url) pathComponents.
	class := pc first asSymbol.
	selector := pc second asSymbol.
	
	Smalltalk at: class ifPresent: [:c | 
		(c class allSelectors includes: selector) ifTrue: [
			result := c perform: selector]].

	^ HttpResponse fromString: result printString.! !

RestHttpHandler subclass: #RestResourceHttpHandler
	instanceVariableNames: 'resources'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Rest-Server'!

!RestResourceHttpHandler methodsFor: 'http-post-helper' stamp: 'mt 2/16/2009 17:18'!
create: resClass from: fields
	"Works only with a new project or a new user."
	
	| entity |
	entity := resClass new.
	fields keysAndValuesDo: [:key :value |
		entity instVarNamed: key asString put: value first].
	resClass perform: ('add', (self resolveName: resClass) capitalized, ':') asSymbol with: entity.
	^ HttpResponse status: #ok contents: entity id asString! !

!RestResourceHttpHandler methodsFor: 'http-delete-helper' stamp: 'mt 2/16/2009 17:36'!
delete: resClass id: id
	"Delete a resource with the specified identifier. Works only with projects and users."

	| entity |
	entity := resClass all
		detect: [:each | each id = id]
		ifNone: [^ self responseForNotFound: resClass id: id].
	resClass
		perform: ('delete', (self resolveName: resClass) capitalized, ':') asSymbol
		with: entity.
	^ HttpResponse status: #ok contents: 'deleted'.! !

!RestResourceHttpHandler methodsFor: 'private' stamp: 'mt 2/16/2009 16:35'!
getRequestChainFor: aPathComponentList
	"Create a list of associations for the request. That list contains resource classes and their requested ids."
	
	| chain resClass idNumber |
	chain := OrderedCollection new.
	aPathComponentList size odd ifTrue: [aPathComponentList add: 'nil'].
	aPathComponentList pairsDo: [:resName :idName | 
		resClass := self resolveResource: resName ifAbsent: [nil].
		idName = 'nil'
			ifTrue: [idNumber := nil]
			ifFalse: [idNumber := idName asNumber].
		resClass ifNotNil: [chain add: (resClass -> idNumber)]].
	^ chain! !

!RestResourceHttpHandler methodsFor: 'initialization' stamp: 'mt 2/16/2009 15:41'!
initialize
	"Register all accessible resources."
	
	super initialize.
	resources := Dictionary new.
	self registerXpfEntityClasses.! !

!RestResourceHttpHandler methodsFor: 'request processing' stamp: 'cs 2/17/2009 18:37'!
process: anHttpRequest 
	"Create a resource chain from the path of the http request. Fail if that chain is empty. Redirect processing according to used verb (POST, PUT, GET, DELETE)."

	| res |
	res := self getRequestChainFor: (URI fromString: anHttpRequest url) pathComponents.
	res isEmpty ifTrue: [^ self responseForFailure: anHttpRequest].
	
	anHttpRequest isPostRequest ifTrue: [^ self processPost: res fields: anHttpRequest postFields].
	anHttpRequest isGetRequest ifTrue: [^ self processGet: res].
	anHttpRequest isPutRequest ifTrue: [^ self processPut: res fields: anHttpRequest putFields].
	anHttpRequest isDeleteRequest ifTrue: [^ self processDelete: res].! !

!RestResourceHttpHandler methodsFor: 'request processing' stamp: 'mt 2/16/2009 16:59'!
processDelete: aResourceChain 

	|resClass id|
	resClass := aResourceChain first key.
	id := aResourceChain first value.
	^ self delete: resClass id: id! !

!RestResourceHttpHandler methodsFor: 'request processing' stamp: 'mt 2/16/2009 16:57'!
processGet: aResourceChain 

	|resClass id|
	resClass := aResourceChain first key.
	id := aResourceChain first value.
	
	id isNil
		ifTrue: [^ self read: resClass]
		ifFalse:[^ self read: resClass id: id].! !

!RestResourceHttpHandler methodsFor: 'request processing' stamp: 'mt 2/16/2009 16:56'!
processPost: aResourceChain fields: someFields 
	
	|resClass|
	resClass := aResourceChain first key.
	^ self create: resClass from: someFields! !

!RestResourceHttpHandler methodsFor: 'request processing' stamp: 'mt 2/16/2009 16:58'!
processPut: aResourceChain fields: someFields 

	|resClass id|
	resClass := aResourceChain first key.
	id := aResourceChain first value.
	^ self update: resClass from: someFields id: id! !

!RestResourceHttpHandler methodsFor: 'http-get-helper' stamp: 'mt 2/16/2009 17:24'!
read: resClass
	"Read information about all entities of a given class. Works only for projects or users."
	
	| response |
	response := resClass all collect: [:entity|
		Dictionary newFrom: (
			resClass allInstVarNames collect: [:name| name -> (entity instVarNamed: name)])].
	^ HttpResponse status: #ok contents: (Json render: response).! !

!RestResourceHttpHandler methodsFor: 'http-get-helper' stamp: 'mt 2/16/2009 17:36'!
read: resClass id: id
	"Read information about one special resource with the given id. Works only for projects and users."

	| entity response |
	entity := resClass all
		detect: [:each | each id = id]
		ifNone: [^ self responseForNotFound: resClass id: id].
	response := Dictionary new.
	XpfProject allInstVarNames do:
		[:name| response at: name put: (entity instVarNamed: name)].
	^ HttpResponse status: #ok contents: (Json render: response).! !

!RestResourceHttpHandler methodsFor: 'initialization' stamp: 'mt 2/16/2009 15:42'!
registerXpfEntityClasses
	"Add all subclasses of XpfEntity to the pool of accessible resources. Example: 'XpfUserStory' will have the key 'userstory' to lookup the entity class."

	XpfEntity allSubclasses do: [:aClass |
		resources
			at: (aClass name copyReplaceAll: 'Xpf' with: '') asLowercase
			put: aClass]! !

!RestResourceHttpHandler methodsFor: 'private' stamp: 'mt 2/16/2009 17:35'!
resolveName: aClass
	"Inverted lookup in the pool of registered resourcen."

	resources keysAndValuesDo: [:key :value |
		value = aClass ifTrue: [^ key]].
	^ nil.! !

!RestResourceHttpHandler methodsFor: 'private' stamp: 'mt 2/16/2009 16:45'!
resolveResource: aResourceName ifAbsent: aBlock
	"Try to resolve the given resource name to a key in the registered resource pool. Plural will be converted to singular: userstories -> userstory, projects -> project. The resource name has to be lowercase."

	| resourceKey |
	resourceKey := aResourceName.
	(resourceKey endsWith: 'ies')
		ifTrue:[resourceKey := (resourceKey truncateTo: resourceKey size - 3), 'y'].
	(resourceKey endsWith: 's')
		ifTrue:[resourceKey := resourceKey truncateTo: resourceKey size - 1].
		
	^ resources at: resourceKey ifAbsent: aBlock.! !

!RestResourceHttpHandler methodsFor: 'private' stamp: 'mt 2/16/2009 17:37'!
responseForFailure: anHttpRequest 
	^ nil asHttpResponseTo: anHttpRequest! !

!RestResourceHttpHandler methodsFor: 'private' stamp: 'mt 2/16/2009 17:35'!
responseForNotFound: resource id: id
	^ HttpResponse
		status: #notFound
		contents:
			(self resolveName: resource capitalized),
			' with id ', id asString,
			' not found.'! !

!RestResourceHttpHandler methodsFor: 'http-put-helper' stamp: 'mt 2/16/2009 17:36'!
update: resClass from: fields id: id
	"Works only with projects and users."

	| entity |
	entity := resClass all
		detect: [:each | each id = id]
		ifNone: [^ self responseForNotFound: resClass id: id].
	fields keysAndValuesDo: [:key :value |
		entity instVarNamed: key asString put: value first].
	^ HttpResponse status: #ok contents: 'updated'! !

!XpfEntity methodsFor: '*rest-server' stamp: 'cs 2/16/2009 00:40'!
jsonWriteOn: aStream
	self id jsonWriteOn: aStream! !

!Dictionary methodsFor: '*rest-server' stamp: 'mt 2/16/2009 10:28'!
keysAndValuesCollect: aBlock 
	| newDict |
	newDict := Dictionary new.
	self keysAndValuesDo: [:key :value | newDict add: (aBlock value: key value: value)].
	^ newDict! !

!Collection methodsFor: '*rest-server' stamp: 'cs 2/16/2009 00:38'!
jsonWriteOn: aStream
	| needComma |
	needComma := false.
	aStream nextPut: $[.
	self do: [:v |
		needComma
			ifTrue: [ aStream nextPutAll: ', ' ]
			ifFalse: [ needComma := true ].
		v jsonWriteOn: aStream.
	].
	aStream nextPut: $].! !

!HttpRequest methodsFor: '*rest-server' stamp: 'cs 2/17/2009 18:34'!
putFields
	"Answer the multipart fields for a post request (if in fact this is a
	POST request"

	^self propertyAt: #putFields ifAbsentPut: [
		" parse POST fields in the body of request, if it's standard form data "

		(self isPutRequest and: [ 
			self contentType = MIMEDocument contentTypeFormData 
		]) ifTrue: [
			self class
				decodeUrlEncodedForm: (stream next: self contentLength)
				multipleValues: true.
		] ifFalse: [
			Dictionary new
		].
	]! !
